""" The RootWindow and Window classes """

import sys
import types
import _curses
curses=_curses
from Rect import *
from NoDebug import *

if 0:
    def wibble(o,i=''):
    #    print 'o=',o,type(o)
        for g in o.keys():
            if type(o[g])==types.DictionaryType:
                print i,'%s={'%g
                wibble(o[g],i+'    ')
                print i,'}'
            else:
                print i,g,'\t\t=',o[g]

    def frog():
        wibble(globals())
    #    wibble(sys.modules)

###############################################################################

class RootWindow:
    """ Proto-window class. You probably only ever want one of these. In fact
        you can only ever have one, as explained by the next sentence.
        Static (in the C++ sense) variable exist here:
            * rootwindow: The most recently created RootWindow.
#            * attrmap: attribute map (maps the 256 combinations of foreground/
#                background colours into ncurses attributes)
        Note that to not make Python confuse static functions with nonstatic
        methods the former is implemented as a tuple containing the function.

        I have yet to decide whether this is an abuse of the language.

        I am thinking of a RootWindow as the location of a master UI lock.

        Remember that curses is cleaned up at the end of the constructor, so
        don't bother trying to do anything after __init__ exits...
    """
    def __init__(self,main):
        """ Initialise ncurses & basic window system, call main(), and then clean up.
        """
        print("Initialising curses")
        self.stdscr=curses.initscr()

        try:
            self._init()
            main(self)
        finally:
            curses.endwin()
#            print("ESC[H\ncurses deinit'd")
            print("\ncurses deinit'd")

    def _init(self):
        """ Protected part of initialisation sequence
        """
        curses.noecho()
        curses.cbreak()
        curses.nonl()

        # first, let's do some sanity checks on the ncurses Python module.

        if not 'bkgd' in dir(self.stdscr):        # really crap cursesmodule
            debugmsg('Ack! Really crappy Python (n)curses module! You must upgrade!')
            sys.exit(1)

        if curses.__dict__.has_key('key_f1'):
            # define NCURSES_VERSION "1.9.9g"
            # colours might also be swapped...

            debugmsg('Ack! Your Python ncurses module contains lower case KEY_* constants (keyname()\'s fault)!')
            debugmsg('Try updating ncurses and/or your Python ncurses module.')
            sys.exit(1)

        if curses.__dict__.has_key('COLOR_PAIR'):
            debugmsg('Your Python ncurses module is not the latest. Upgrade recommended.')
            curses.color_pair=curses.COLOR_PAIR;

        # mermaid : curses.h.in,v 1.36 1996/11/16 21:55:54 tom
        # mussel : curses.h.in,v 1.60 1998/02/11 12:13:46 tom

        # If we have colour capability:
        if curses.__dict__.has_key('has_colors') and curses.has_colors():
            curses.start_color()

            # init 64 colour pairs
            colour=[curses.COLOR_BLACK,curses.COLOR_BLUE,curses.COLOR_GREEN,curses.COLOR_CYAN,
                    curses.COLOR_RED,curses.COLOR_MAGENTA,curses.COLOR_YELLOW,curses.COLOR_WHITE]
            for index in range(1,64):
                # The ordering of colours means that index 0 has the colour we
                # wanted to (white on black), because we aren't able to change
                # it.
                curses.init_pair(index,colour[index&7^7],colour[index/8])

            def funkify(index):
                """ map (colour) index 0..255 to curses Attribute (colour version) """
                return curses.A_ALTCHARSET*0+(index&128!=0)*curses.A_BLINK+ \
                       (index&8!=0)*curses.A_BOLD+    \
                       curses.color_pair((index&112)/2+(index&7^7))
        else:
            def funkify(index):
                """ map (colour) index 0..255 to curses Attribute (mono version) """
                # This may need a bit of tweaking.
                return (index&128!=0)*curses.A_CHARTEXT*0+ \
                       (index&8!=0)*curses.A_BOLD

        # map fg/bg colour combinations to attributes (tricky on mono terminals! :)
        self.attrmap=map(funkify,range(0,256))
        RootWindow.rootwindow=(self,)    # magic here

        self.windows=[self]              # windows[-1] will always be the root window. [0] is the top window
        self.root=self                   # I *am* the root window.
        self.hgt,self.wid=self.stdscr.getmaxyx() # not sure if this is required.
        self.rect=BaseRect(self.wid,self.hgt)
        self.pad=curses.newpad(self.hgt,self.wid)
        # unfortunately not every CP437 graphics character has an ACS_*
        # representation.
        self.pad.bkgd(self.attrmap[0x1E]+ord(''))
#        self.pad.bkgd(self.attrmap[0x1E]+curses.ACS_BOARD)
        self.update()

    def __repr__(self):
        return '<RootWindow %dx%d>'%(self.wid,self.hgt)

    def update(self,rect=None):
        """ noutrefresh & update visible portion of rect
            rect is relative to its parent.
        """
        windows=self.root.windows

#        debugmsg('\nRootWindow.update(): updating %s'%self)
        if not rect:
            rect=Rect(self.rect,(0,0,0,0,0,0,1,1)) # this is effectively a copy constructor...
#            debugmsg('    RootWindow.update(): defaulting rect to %s'%rect)

        # check to make sure window is visible
        if not self in windows:          # window not visible?
#            debugmsg('    RootWindow.update(): window is hidden, nothing to do!')
            return                       # nothing to do

        wpos=[rect.gxyxy_clip()]         # convert to screen coords & clip to self and super-rects
        if not wpos[0]:
#            debugmsg('RootWindow.update(): %s is totally off screen'%rect)
##            debugmsg('RootWindow.update(): continuing anyway!')
            return                       # completely off screen

#        debugmsg('RootWindow.update(): preclip wpos=%s'%wpos)
        for win in windows[0:windows.index(self)]: # for all windows above rect's window
            wpos=win.rect.mask_out(wpos)
#        debugmsg('RootWindow.update(): postclip wpos=%s'%wpos)

        for r in wpos:
#            debugmsg('    r=%s (this (partial) rect)'%(r))
##            debugmsg('    r.add=%s'%(r.add))
            w=r.add(-self.rect.x1(),-self.rect.y1())       # position of redrawn portion of rect relative to window
#            debugmsg('    about to update %s->%s'%(r,w))
            self.pad.noutrefresh(w.y1(),w.x1(),r.y1(),r.x1(),r.y2()-1,r.x2()-1)
        curses.doupdate()

###############################################################################

class Window(RootWindow):
    """ A basic window. Widgets sold separately.
    """
    def __init__(self,parent,pos,colour=None,border_colour=None):
        """
        pos is relative to the parent's position
        It is an 8-record tuple.
        """
#        self.pos=parent.rect(pos)
        self.rect=Rect(parent.rect,pos)
        self.parent=parent
        self.root=parent.root            # if the parent *is* the root then parent==root
        self.pad=curses.newpad(self.rect.hgt(),self.rect.wid())
        self.colour={}
        if not colour:
            colour=0x07
        self.colour['window']=colour
        if not border_colour:
            border_colour=0x0F
        self.colour['border']=colour
#        self.set_bgcol(border_colour)
        self.pad.bkgd(self.root.attrmap[border_colour]+255)
        self.pad.box()
#        self.set_bgcol(colour)
        self.pad.nodelay(1)
        self.pad.keypad(1)
#        self.taborder=[]                 # focusable widgets, in order
#        self.curr_widget=None
        self.widgets=[]                  # all widgets
        self.is_visible=0

    def __getitem__(self,item):          # this is just a convenience of notation
        return self.widget[item]         #   (ie. somewindow.widget["somewidget"] -> somewindow[somewidget])

    def __repr__(self):
        return '<Window at %s>'%(self.rect)

#    def __del__(self):                   # not sure if this is actually at all useful
#        self.hide()                      # this could cause nasty recursion...

    def show(self):
        if not self.is_visible:
            self.root.windows[0:0]=[self] # insert self at beginning of list
            self.is_visible=1
            self.update()                # redraw all of self

    def hide(self):
#        # if visible or either doesn't have a handle_close() or it returns non-zero:
#        if self.is_visible and (not hasattr(self,'handle_close') or self.handle_close()):

        if self.is_visible: # (re)introduce allow_close() at some stage
            self.is_visible=0
            self.root.windows.remove(self)
            self.root.update(self.rect)          # redraw space previously occupied by window...UNTESTED!

    def move(self,pos):
        """ Move and/or resize window to new location pos
            pos is an 8(or 4)-tuple
        """
        old_rect=self.rect.copy()
#        self.rect=Rect(self.parent.rect,pos)
        self.rect.set(pos)
        if (old_rect.wid(),old_rect.hgt())!=(self.rect.wid(),self.rect.hgt()): # size changed?
            # I should do something about the old pad...
            debugmsg('Window.move(): size changed from %s to %s'%((old_rect.wid(),old_rect.hgt()),(self.rect.wid(),self.rect.hgt())))
            if self.rect.hgt()==0 and self.rect.wid()==0:
                raise "","Whoops, null window: "+str(self.rect)
            self.pad=curses.newpad(self.rect.hgt(),self.rect.wid())
            self.pad.bkgd(self.root.attrmap[self.colour['border']]+255)
            self.pad.box()
            self.pad.nodelay(1)
            self.pad.keypad(1)
            debugmsg('Window.move(): widgets=%s'%(self.widgets))
            for w in self.widgets:
                w.init()
#                w.draw()
        debugmsg('\nWindow.move(): About to update '+str(self.rect))
        self.update(self.rect)
        debugmsg('Window.move(): About to update '+str(old_rect))
        self.root.update(old_rect)
#        self.root.update()

